#include "prjlibs-c/standards.h"
#include <sys/stat.h>
#include <unistd.h>
#include <limits.h>
#include <time.h>
#include <errno.h>

#include "skalibs/include/stddjb.h"
#include "prjlibs-c/constants.h"
#include "prjlibs-c/nsec.h"

char const* PROG="statfile";

static void set(char const* const var, char const* const val) {
  if (pathexec_env(var+1, val)==0)
    strerr_die5sys(111, PROG, err_unable, err_setvar, var, ": ");
}

static void set64n(char const* const var, time_t t, unsigned int nsec) {
  unsigned char packed[TAI_PACK+4];
  {
    struct tai stamp;
    tai_unix(&stamp, t);
    tai_pack((char*)packed, &stamp);
  }
  packed[TAI_PACK+3]=nsec%256; nsec/=256;
  packed[TAI_PACK+2]=nsec%256; nsec/=256;
  packed[TAI_PACK+1]=nsec%256; nsec/=256;
  packed[TAI_PACK+0]=nsec%256;
  {
    static char const hex[16]="0123456789abcdef";
    char buf[(2*sizeof packed)+1];
    unsigned int i;
    for (i=0; i<sizeof packed; ++i) {
      buf[2*i+0]=hex[packed[i]/16];
      buf[2*i+1]=hex[packed[i]%16];
    }
    buf[(sizeof buf)-1]='\0';
    set(var, buf);
  }
}

int main(int argc, char** argv) {
  struct stat statbuf;
  char buf[((sizeof (unsigned long))*CHAR_BIT+2)/3+1];
  int use_lstat  =0;
  int set_dev    =0;
  int set_ino    =0;
  int set_type   =0;
  int set_mode   =0;
  int set_nlink  =0;
  int set_uid    =0;
  int set_gid    =0;
  int set_rdev   =0;
  int set_size   =0;
  int set_blksize=0;
  int set_blocks =0;
  int set_atime  =0;
  int set_mtime  =0;
  int set_ctime  =0;

  if (argc<3)
    strerr_die3x(100, usage, PROG, " -lditmnugrsSbAMC file [command arg ...]");

  {
    char const* x=argv[1];
    if (*x=='-') ++x;
    while (*x!='\0') {
      switch (*x) {
        case 'l': use_lstat  =1; break;
        case 'd': set_dev    =1; break;
        case 'i': set_ino    =1; break;
        case 't': set_type   =1; break;
        case 'm': set_mode   =1; break;
        case 'n': set_nlink  =1; break;
        case 'u': set_uid    =1; break;
        case 'g': set_gid    =1; break;
        case 'r': set_rdev   =1; break;
        case 's': set_size   =1; break;
        case 'S': set_blksize=1; break;
        case 'b': set_blocks =1; break;
        case 'A': set_atime  =1; break;
        case 'M': set_mtime  =1; break;
        case 'C': set_ctime  =1; break;
        default: {
          char opt[2]={ '\0', '\0' };
          opt[0]=*x;
          strerr_die3x(100, PROG, "unrecognized option: ", opt);
        }
      }
      ++x;
    }
  }

  {
    int ret;
    if ('0'<=argv[2][0] && argv[2][0]<='9') {
      unsigned int fd;
      unsigned int const len=scan_uint(argv[2], &fd);
      if (len==0 || argv[2][len]!='\0')
        strerr_die3x(100, PROG, "fd malformed: ", argv[2]);
      if (fd>INT_MAX)
        strerr_die2x(100, PROG, "fd overflows int");
      ret=fstat(fd, &statbuf);
    } else
      ret=(use_lstat? lstat: stat)(argv[2], &statbuf);
    if (ret!=0) {
      ret=errstat;
      if (argc==3) _exit(ret);
      strerr_die5sys(ret, PROG, err_unable, err_stat, argv[2], ": ");
    }
  }

  if (argc==3) _exit(0);

  #define doit(member, var, type) do { \
    if (statbuf.st_##member<0 || \
        statbuf.st_##member!=(type)(unsigned long)statbuf.st_##member) \
      strerr_die4x(100, PROG, #member, err_oflow, "unsigned long"); \
    if (set_##member) { \
      buf[fmt_ulong(buf, statbuf.st_##member)]='\0'; \
      set(var, buf); \
    } \
  } while (0)

  doit(dev,     "$DEV",     dev_t);
  doit(ino,     "$INO",     ino_t);
  doit(nlink,   "$NLINK",   nlink_t);
  doit(uid,     "$UID",     uid_t);
  doit(gid,     "$GID",     gid_t);
  doit(rdev,    "$RDEV",    dev_t);
  doit(size,    "$SIZE",    off_t);
  doit(blksize, "$BLKSIZE", unsigned long);
  doit(blocks,  "$BLOCKS",  unsigned long);
  doit(atime,   "$ATIME",   time_t);
  doit(mtime,   "$MTIME",   time_t);
  doit(ctime,   "$CTIME",   time_t);

  if (set_atime) set64n("$ATAI64N", statbuf.st_atime, atime_nsec(&statbuf));
  if (set_mtime) set64n("$MTAI64N", statbuf.st_mtime, mtime_nsec(&statbuf));
  if (set_ctime) set64n("$CTAI64N", statbuf.st_ctime, ctime_nsec(&statbuf));

  if (set_type) {
    char const* type;
    #ifdef S_ISREG
      if (S_ISREG(statbuf.st_mode)) type="-"; else
    #endif
    #ifdef S_ISDIR
      if (S_ISDIR(statbuf.st_mode)) type="d"; else
    #endif
    #ifdef S_ISLNK
      if (S_ISLNK(statbuf.st_mode)) type="l"; else
    #endif
    #ifdef S_ISCHR
      if (S_ISCHR(statbuf.st_mode)) type="c"; else
    #endif
    #ifdef S_ISBLK
      if (S_ISBLK(statbuf.st_mode)) type="b"; else
    #endif
    #ifdef S_ISFIFO
      if (S_ISFIFO(statbuf.st_mode)) type="p"; else
    #endif
    #ifdef S_ISSOCK
      if (S_ISSOCK(statbuf.st_mode)) type="s"; else
    #endif
    type="unknown";
    set("$TYPE", type);
  }

  if (set_mode) {
    unsigned long const mode=statbuf.st_mode&07777;
    buf[0]='0';
    buf[1]='0'+((mode>>9)&07);
    buf[2]='0'+((mode>>6)&07);
    buf[3]='0'+((mode>>3)&07);
    buf[4]='0'+(mode&07);
    buf[5]='\0';
    set("$MODE", buf);
    set("$MODE_SUID",   (statbuf.st_mode&04000? "1": null));
    set("$MODE_SGID",   (statbuf.st_mode&02000? "1": null));
    set("$MODE_STICKY", (statbuf.st_mode&01000? "1": null));
    set("$MODE_RUSR",   (statbuf.st_mode&00400? "1": null));
    set("$MODE_WUSR",   (statbuf.st_mode&00200? "1": null));
    set("$MODE_XUSR",   (statbuf.st_mode&00100? "1": null));
    set("$MODE_RGRP",   (statbuf.st_mode&00040? "1": null));
    set("$MODE_WGRP",   (statbuf.st_mode&00020? "1": null));
    set("$MODE_XGRP",   (statbuf.st_mode&00010? "1": null));
    set("$MODE_ROTH",   (statbuf.st_mode&00004? "1": null));
    set("$MODE_WOTH",   (statbuf.st_mode&00002? "1": null));
    set("$MODE_XOTH",   (statbuf.st_mode&00001? "1": null));
  }

  argv+=3;
  pathexec((char const**)argv);
  strerr_die5sys(errstat, PROG, err_unable, err_exec, argv[0], ": ");
}
